""" Function to be imported in a WaterFrame. It save the WaterFrame into a NetCDF file. """
from xarray import Dataset


def to_nc(self, path=None, nc_format="NETCDF4"):
    """
    It saves the WaterFrame into a NetCDF.

    Parameters
    ----------
        path: str
            Path to save the NetCDF. If path is None, the filename will be metadata['id'].
        nc_format: str
            Specify the NetCDF format (NETCDF3_64BIT, NETCDF4).

    Returns
    -------
        path: str
            Path where the file is placed.
    """

    if path is None:
        path = self.metadata['id'] + '.nc'

    # Sometimes the metadata contains a list of str.
    # It can crash the creation of the NetCDF3_64Bits.
    # We are going to change the list of str to str
    metadata_netcdf = self.metadata.copy()
    for key, value in metadata_netcdf.items():
        if isinstance(value, list):
            metadata_netcdf[key] = ', '.join(value)

    # Multiindex is not alloud yet in to_netcdf of xarray
    df = self.data.reset_index()
    df.set_index('TIME', inplace=True)

    # Creation of an xarray dataset
    ds = Dataset(data_vars=df, attrs=metadata_netcdf)
    for key in self.parameters:
        try:
            ds[key].attrs = self.vocabulary[key]
        except KeyError:  # Variable without vocabulary
            pass

    # Creation of the nc file
    ds.to_netcdf(path, format=nc_format)

    return path



""" Implementation of mooda """

from .waterframe import WaterFrame
from .input import read_nc_emodnet, read_nc, read_nc_imos, read_pkl, \
    read_nc_moist, from_emso
from .util import concat, iplot_location, iplot_timeseries, md5, \
    es_create_indexes, EMSO, widget_qc, widget_emso, widget_emso_qc, widget_save

__version__ = '1.9.0'



""" Input functions """
from .read_nc_emodnet import read_nc_emodnet
from .read_nc import read_nc
from .read_nc_ooi import read_nc_ooi
from .read_nc_imos import read_nc_imos
from .read_pkl import read_pkl
from .read_json import read_json
from .read_nc_moist import read_nc_moist
from .from_emso import from_emso



from typing import List
import pandas as pd
from pandas.core.frame import DataFrame
from ..waterframe import WaterFrame
from ..util import EMSO

def from_emso(platform_code: str, parameters: List[str]=[], start_time: str='',
              end_time: str='', depth_min: float=None, depth_max: float=None,
              user: str='', password: str='', size: int=10, token: str=''
              ) -> WaterFrame:
    """
    Get a WaterFrame with the data of the EMSO API (api.emso.eu).

    Parameters
    ----------
        user: str
            Login for the EMSO ERIC API.
        password: str
            Password for the EMSO ERIC API.
        token: str
            Token for the EMSO ERIC API.
        platform_code: str
            Data filtered by platform_code
        parameters: List[str]
            List of parameters to get data
        start_time: str
            First date of the measurement
        end_time: str
            Last date of the measurement
        depth_min: float
            Minimum depth of the measurement
        depth_max: float
            Maximum depth of the measurement
        size: int
            Number of values
    
    Returns
    -------
        wf: WaterFrame object
    """
    # Login to EMSO
    emso = EMSO(user=user, password=password, token=token)

    # Get metadata    
    metadata_list = emso.get_metadata(platform_codes=[platform_code])

    # Get data
    complete_data = DataFrame()
    for parameter in parameters:
        data_list = emso.get_data(platform_codes=[platform_code], parameters=[parameter],
                                  start_time=start_time, end_time=end_time, depth_min=depth_min,
                                  depth_max=depth_max, size=size, sort='asc')

        if data_list:
            data = pd.DataFrame(data_list)

            data = data.rename(
                columns = {
                    'time': 'TIME',
                    'time_qc': 'TIME_QC',
                    'value': parameter,
                    'value_qc': f'{parameter}_QC',
                    'depth': 'DEPTH',
                    'depth_qc': 'DEPTH_QC' 
                })

            del data['parameter']
            del data['metadata_id']
            del data['platform_code']
            del data['institution']
            del data['area']
            del data['long_name']
            del data['units']
            del data['location']
            del data['location_qc']

            data['TIME'] = pd.to_datetime(data['TIME'])
            data.set_index(['DEPTH', 'TIME'], inplace=True)
            data = data.astype(float)
            complete_data = complete_data.append(data)

    wf = WaterFrame()
    if metadata_list:
        wf.metadata = metadata_list[0]
    wf.data = complete_data.copy()
    for parameter in wf.parameters:
        find = False
        for metadata_parameter in wf.metadata['parameters']:
            try:
                acronym = metadata_parameter.split('-')[0].strip()
                long_name = metadata_parameter.split('-')[1].strip().split('[')[0].strip()
                units = metadata_parameter.split('-')[1].strip().split('[')[1][:-1]
                if parameter == acronym:
                    wf.vocabulary[acronym] = {'long_name': long_name, 'units': units}
                    find = True
                    break
            except:
                pass
        if not find:
            wf.vocabulary[parameter] = {'long_name': '', 'units': ''}

    return wf



""" Implementation of util """
from .concat import concat
from .iplot import iplot_location, iplot_timeseries
from .md5 import md5
from .es_create_indexes import es_create_indexes
from .emso import EMSO
from .md_widgets import widget_qc, widget_emso, widget_emso_qc, widget_save



"""
Function to be imported in a WaterFrame. It creates a JSON with the WaterFrame information.
"""
import json


def to_json(self):
    """
    Get a JSON with the WaterFrame information.

    Returns
    -------
        json_string: str
            JSON string.
    """
    # Convert all dict values into str
    big_dict = {
        "metadata": dict(zip(self.metadata, map(str, self.metadata.values()))),
        "vocabulary": dict(zip(self.vocabulary, map(str, self.vocabulary.values()))),
        "data": self.data.to_json()
    }

    json_string = json.dumps(big_dict)

    return json_string



"""
Implementation of a function to read NetCDF files from Integrated Marine Observing System (IMOS)
"""
import xarray as xr
from ..waterframe import WaterFrame


def read_nc_imos(path, clean_data=True):
    """
    Open a NetCDF file from Integrated Marine Observing System (IMOS). All QC flags are 0.

    Parameters
    ----------
        path: str
            Path of the NetCDF file.
        clean_data: bool
            It deletes variables with _burst_min, _burst_max, _burst_sd, _num_obs, TIMESERIES and
            NOMINAL_DEPTH in name.

    Returns
    -------
        wf: WaterFrame
    """
    def drop(ds_in, clean_data_in):
        """Drop some parameters of the dataset.

        Parameters
        ----------
            ds_in: xarray.Dataset
                Input dataset.
            clean_data_in: bool
                It indicates if "_DM" variables should be deleted.
        Returns
        -------
            ds_out: xarray.Dataset
                Output dataset
        """

        ds_out = ds_in
        if clean_data_in:

            for key in ds_in.variables.keys():
                # This id / elif / elif can be just one if... I know...
                if key == 'TIMESERIES' or key == 'NOMINAL_DEPTH':
                    ds_out = ds_out.drop(key)
                elif '_burst_min' in key or '_burst_max' in key or '_burst_sd' in key:
                    ds_out = ds_out.drop(key)
                elif '_num_obs' in key:
                    ds_out = ds_out.drop(key)

        return ds_out

    # Create WaterFrame
    wf = WaterFrame()

    # Open file with xarrat
    ds = xr.open_dataset(path)
    ds = drop(ds, clean_data)
    # Save ds into a WaterFrame
    wf.metadata = dict(ds.attrs)
    wf.data = ds.to_dataframe()
    for key in wf.data.keys():
        wf.data[f'{key}_QC'] = 0
    for variable in ds.variables:
        wf.vocabulary[variable] = dict(ds[variable].attrs)

    return wf



""" Implementation of mooda.widget_save()"""
import ipywidgets as widgets


def widget_save(wf):
    """
    Make a Jupyter notebook widget that allows to save the WaterFrame to a file

    Parameters
    ----------
        wf: WaterFrame
    
    Returns
    -------
        main_box: ipwidgets.VBox
            Jupyter notebook widget 
    """
    # Title
    title_label = widgets.Label('Save as...')

    # Filename
    name_label = widgets.Label('Name:')
    input_name = widgets.Text()

    # Type
    type_label = widgets.Label('Type:')
    input_type = widgets.RadioButtons(
        options=['csv', 'nc', 'pkl'], value='csv', disabled=False)

    # Button
    button = widgets.Button(description='Save')
    out_save = widgets.Output()
    def on_button_clicked(_):
        with out_save:
            if input_name.value:
                if input_type.value == 'csv':
                    wf.to_csv(path=f'{input_name.value}.csv')
                elif input_type.value == 'nc':
                    wf.to_nc(path=f'{input_name.value}.nc')
                elif input_type.value == 'pkl':
                    wf.to_pkl(path=f'{input_name.value}.pkl')
                print('Done.')
            else:
                print('Please, write a name.')
    # linking button and function together using a button's method
    button.on_click(on_button_clicked)

    filename_box = widgets.HBox([name_label, input_name])
    type_box = widgets.HBox([type_label, input_type])

    main_box = widgets.VBox([title_label, filename_box, type_box, button,
                             out_save])

    return main_box



""" Implementation of wf.qc_syntax_test() """

def qc_syntax_test(self):
    """
    It verifies that wf.data contains the recomentated data structure to pass
    the rest of the QC test. Each parameter (for example, TEMP) and each index
    (for example, TIME) must have a column with the flags of the values
    (for example, TEMP_QC and TIME_QC).

    Returns
    -------
        success: bool
            The test has been successfully passed.
    """
    # Get all columns and indices
    all_columns = list(self.data.keys())
    all_columns += list(self.data.index.names)
    # Get only the names without _QC
    columns_no_qc = [column_name for column_name in all_columns if '_QC' not in column_name]

    success = True
    
    for column_no_qc in columns_no_qc:
        if f'{column_no_qc}_QC' not in all_columns:
            success = False

    return success



""" Implementation of WaterFrame.copy() """
from copy import deepcopy


def copy(self):
    """
    It returns a copy of the WaterFrame.

    Returns
    -------
        new_wf: WaterFrame
            A copy of the WaterFrame.
    """
    new_wf = deepcopy(self)
    return new_wf



""" Implementation of mooda.widget_emso()"""
from typing import List
from IPython.core.display import display
import ipywidgets as widgets
from IPython.display import clear_output
from datetime import date
import plotly.offline as pyo
import plotly.io as pio
import mooda


def widget_emso_qc(wf, depth_range: List[float]=[-10, 10000],
                   range_test: List[float]=[-1000, 1000],
                   spike_window: int=100, spike_threshold: float=3.5,
                   spike_influence: float=0.5, user: str='',
                   password: str='', token: str=''):
    """
    It makes a Jupyter Notebook Widget used to download EMSO data and make data
    quality control tests.

    Parameters
    ----------
        wf: WaterFrame
        depth_range: List[float]
            Range of depth
        user: str
            User for the EMSO ERIC API
        password: str
            Password for the EMSO ERIC API
        token: str
            Token for the EMSO ERIC API
        parameter: str
            Parameter
        range_test: List[float]
            Limits for the range test [min value, max value]
        spike_window: int
            Window for the spike test
        spike_threshold: float
            Float for the spike test
        spike_influence: float
            Influence of the bad data on the spike test

    Returns
    -------
        main_box: ipwidgets.VBox
            Jupyter notebook widget 
    """
    pyo.init_notebook_mode()

    def show_result(wf, parameter_in, chart_title=''):
        # Change name of flags
        wf2 = wf.copy()
        qc_labels = {0: 'No QC', 1: 'Good data', 4: 'Bad data'}
        wf2.data[f'{parameter_in}_QC'].replace(qc_labels, inplace=True)

        fig = wf2.iplot_line(
            parameter_in,
            color=f'{parameter_in}_QC',
            marginal_y=None,
            line_shape='linear',
            rangeslider_visible=False,
            line_dash_sequence=['dot', 'dot'],
            title=chart_title)
        fig.show()

    # Sing in
    user_label = widgets.Label('User:')
    input_user = widgets.Text(value=user)
    password_label = widgets.Label('Password:')
    input_password = widgets.Text(value=password)
    token_label = widgets.Label('Token:')
    input_token = widgets.Text(value=token)

    # map
    emso = mooda.EMSO(user=input_user.value, password=input_password.value,
                      token=input_token.value)
    map_dict = emso.get_fig_map()
    pio.show(map_dict)

    # Platform code
    platform_label = widgets.Label('Platform code:')
    platform_codes = emso.get_info_platform_code()
    input_platform = widgets.Dropdown(options=platform_codes,
                                      value=platform_codes[0], disabled=False)
    
    # Get metadata
    metadata_list = emso.get_metadata(platform_codes=[platform_codes[0]])
    try:
        metadata = metadata_list[0]
        parameter_list = metadata.get('parameters')
        if parameter_list is None:
            parameter_list = emso.get_info_parameter()
        depth_min = metadata.get('geospatial_vertical_min', depth_range[0])
        depth_max = metadata.get('geospatial_vertical_max', depth_range[1])
        start_date = metadata.get('time_coverage_start')
        if start_date:
            start_date = start_date.split('T')[0]
        end_date = metadata.get('time_coverage_end')
        if end_date:
            end_date = end_date.split('T')[0]
    except:
        parameter_list = emso.get_info_parameter()
        depth_min = depth_range[0]
        depth_max = depth_range[1]
        start_date = None
        end_date = None

    # Parameters
    parameters_label = widgets.Label('Parameters:')
    input_parameters = widgets.SelectMultiple(options=parameter_list, rows=10,
                                              disabled=False,
                                              value=[parameter_list[0]])

    # Size
    size_label = widgets.Label('Size:')
    input_size = widgets.BoundedIntText(value=10, min=1, step=1)

    # Depth
    depth_label = widgets.Label('Depth:')
    input_depth = widgets.FloatRangeSlider(value=[depth_min, depth_max],
                                           min=depth_min,
                                           max=depth_max, step=0.1,
                                           disabled=False,
                                           continuous_update=False,
                                           orientation='horizontal',
                                           readout=True, readout_format='.1f')

    # Time
    start_date_label = widgets.Label('Start date:')
    input_start_date = widgets.DatePicker(disabled=False,
                                          min_date=date(
                                              int(start_date.split('-')[0]),
                                              int(start_date.split('-')[1]),
                                              int(start_date.split('-')[2])),
                                          max_date=date(
                                              int(end_date.split('-')[0]),
                                              int(end_date.split('-')[1]),
                                              int(end_date.split('-')[2])),
                                          value=date(
                                              int(start_date.split('-')[0]),
                                              int(start_date.split('-')[1]),
                                              int(start_date.split('-')[2])))

    end_date_label = widgets.Label('End date:')
    input_end_date = widgets.DatePicker(disabled=False,
                                        min_date=date(
                                            int(start_date.split('-')[0]),
                                            int(start_date.split('-')[1]),
                                            int(start_date.split('-')[2])),
                                        max_date=date(
                                            int(end_date.split('-')[0]),
                                            int(end_date.split('-')[1]),
                                            int(end_date.split('-')[2])),
                                        value=date(
                                            int(end_date.split('-')[0]),
                                            int(end_date.split('-')[1]),
                                            int(end_date.split('-')[2])))

    qc_label = widgets.Label('QC Tests')

    # Reset flags
    reset_label = widgets.Label('Reset flags:')
    reset_checkbox = widgets.Checkbox(description='Do it!')

    # Flat test
    flat_label = widgets.Label('Flat test:')
    window_flat = widgets.IntText(value=2,
                                    description='Window:',
                                    disabled=False,
                                    color='black')
    flat_checkbox = widgets.Checkbox(description='Do it!')

    # Range test
    range_label = widgets.Label('Range test:')
    range_checkbox = widgets.Checkbox(description='Do it!')
    limits = widgets.FloatRangeSlider(value=[range_test[0], range_test[1]],
                                        min=range_test[0],
                                        max=range_test[1],
                                        step=0.1,
                                        description='Limits:',
                                        disabled=False,
                                        continuous_update=False,
                                        orientation='horizontal',
                                        readout=True,
                                        readout_format='.1f')

    # Spike test
    spike_label = widgets.Label('Spike test:')
    spike_checkbox = widgets.Checkbox(description='Do it!')
    window_spike = widgets.IntText(value=spike_window,
                                    description='Window:', disabled=False,
                                    color='black')
    threshold = widgets.FloatText(value=spike_threshold,
                                    description='Threshold:', disabled=False,
                                    color='black')
    influence = widgets.FloatSlider(
        value=spike_influence,
        min=0,
        max=1,
        step=0.1,
        description='Influence:',
        disabled=False,
        continuous_update=False,
        orientation='horizontal',
        readout=True,
        readout_format='.1f',
        slider_color='white'
    )

    # Replace
    replace_label = widgets.Label('Replace QC:')
    replace_checkbox = widgets.Checkbox(description='Do it!')

    def update_components(_):
        # Update all components
        emso = mooda.EMSO(user=input_user.value, password=input_password.value,
                          token=input_token.value)
        # Get metadata
        metadata_list = emso.get_metadata(platform_codes=[input_platform.value])

        try:
            metadata = metadata_list[0]
            parameter_list = metadata.get('parameters')
            if parameter_list is None:
                parameter_list = emso.get_info_parameter()
            depth_min = metadata.get('geospatial_vertical_min', depth_range[0])
            depth_max = metadata.get('geospatial_vertical_max', depth_range[1])
            start_date = metadata.get('time_coverage_start')
            if start_date:
                start_date = start_date.split('T')[0]
            end_date = metadata.get('time_coverage_end')
            if end_date:
                end_date = end_date.split('T')[0]
        except:
            parameter_list = emso.get_info_parameter()
            depth_min = depth_range[0]
            depth_max = depth_range[1]
            start_date = None
            end_date = None
        
        # Parameters
        input_parameters.options = parameter_list

        # Depth
        input_depth.min = depth_min
        input_depth.max = depth_max
        input_depth.value = [depth_min, depth_max]

        # Time
        input_start_date.min_date = date(int(start_date.split('-')[0]),
                                         int(start_date.split('-')[1]),
                                         int(start_date.split('-')[2]))
        input_start_date.max_date = date(int(end_date.split('-')[0]),
                                         int(end_date.split('-')[1]),
                                         int(end_date.split('-')[2]))
        input_start_date.value = date(int(start_date.split('-')[0]),
                                      int(start_date.split('-')[1]),
                                      int(start_date.split('-')[2]))
        input_end_date.min_date = date(int(start_date.split('-')[0]),
                                       int(start_date.split('-')[1]),
                                       int(start_date.split('-')[2]))
        input_end_date.max_date = date(int(end_date.split('-')[0]),
                                       int(end_date.split('-')[1]),
                                       int(end_date.split('-')[2]))
        input_end_date.value = date(int(end_date.split('-')[0]),
                                    int(end_date.split('-')[1]),
                                    int(end_date.split('-')[2]))

    input_platform.observe(update_components, names='value')

    # Button
    button = widgets.Button(description='Get data')
    out = widgets.Output()
    out2 = widgets.Output()
    def on_button_clicked(_):
        # "linking function with output"
        with out:
            # what happens when we press the button
            clear_output()

            parameters = []
            for input_parameter in input_parameters.value:
                parameters.append(input_parameter.split('-')[0].strip())

            if input_start_date.value is None:
                start_time = ''
            else:
                start_time = str(input_start_date.value) + ' 00:00:00'

            if input_end_date.value is None:
                end_time = ''
            else:
                end_time = str(input_end_date.value) + ' 00:00:00'
            
            if input_token.value is None:
                token = ''
            else:
                token = input_token.value

            print('Downloading data, please wait')

            wf2 = mooda.from_emso(platform_code=input_platform.value,
                                  parameters=parameters, start_time=start_time,
                                  end_time=end_time,
                                  depth_min=input_depth.value[0],
                                  depth_max=input_depth.value[1],
                                  user=input_user.value,
                                  password=input_password.value,
                                  size=input_size.value, token=token)
            

            wf.data = wf2.data.copy()
            wf.metadata = wf2.metadata.copy()
            wf.vocabulary = wf2.vocabulary.copy()

            clear_output()
            display(wf)

        with out2:
            # what happens when we press the button
            clear_output()
            print('Making tests')
            
            for parameter_one in wf.parameters: 

                if reset_checkbox.value:
                    wf.qc_replace(parameters=[parameter_one], to_replace=1,
                                  value=0)
                    wf.qc_replace(parameters=[parameter_one], to_replace=4,
                                  value=0)
                if flat_checkbox.value:
                    wf.qc_flat_test(window=window_flat.value,
                                    parameters=[parameter_one])
                if range_checkbox.value:
                    wf.qc_range_test(limits=limits.value,
                                    parameters=[parameter_one])
                if spike_checkbox.value:
                    wf.qc_spike_test(window=window_spike.value,
                                        influence=influence.value,
                                        threshold=threshold.value,
                                        parameters=[parameter_one])
                if replace_checkbox.value:
                    start = 0
                    if spike_checkbox.value:
                        start = window_spike.value
                    wf.qc_replace(parameters=[parameter_one], start=start)

                show_result(wf, parameter_one, f'Final result')

    # linking button and function together using a button's method
    button.on_click(on_button_clicked)

    user_box = widgets.HBox([user_label, input_user, password_label,
                             input_password])
    token_box = widgets.HBox([token_label, input_token])
    platform_box = widgets.HBox([platform_label, input_platform])
    parameters_box = widgets.HBox([parameters_label, input_parameters])
    size_box = widgets.HBox([size_label, input_size])
    depth_box = widgets.HBox([depth_label, input_depth])
    start_date_box = widgets.HBox([start_date_label, input_start_date])
    end_date_box = widgets.HBox([end_date_label, input_end_date])
    reset_box = widgets.HBox([reset_label, reset_checkbox])
    flat_box = widgets.HBox([flat_label, flat_checkbox, window_flat])
    range_box = widgets.HBox([range_label, range_checkbox, limits])
    spike_column = widgets.VBox([window_spike, threshold, influence])
    spike_box = widgets.HBox([spike_label, spike_checkbox, spike_column])
    replace_box = widgets.HBox([replace_label, replace_checkbox])

    main_box = widgets.VBox([user_box, token_box, platform_box, parameters_box,
                             size_box, depth_box, start_date_box, end_date_box,
                             qc_label, reset_box, flat_box, range_box,
                             spike_box, replace_box, button, out, out2])

    return main_box



""" Implementation of WaterFrame.max_diff(parameter1, parameter2) """


def max_diff(self, parameter1, parameter2):
    """
    It calculates the maximum difference between the values of two parameters.

    Parameters
    ----------
        parameter1: str
            Key name of column 1 to calculate the difference.
        parameter2: str
            Key name of column 2 to calculate the difference.

    Returns
    -------
        where: index
            The position (index) of WaterFrame.data.
        value: float
            Value of the maximum difference.
    """

    where = (self.data[parameter1] - self.data[parameter2]).abs().idxmax()
    value = (self.data[parameter1] - self.data[parameter2]).abs().max()

    return (where, value)



""" Implementation of md.md5 """
import hashlib


def md5(file_path, save_md5=True, md5_path=None):
    """
    It generates the MD5 code of the input file.
    It saves the code into a file if it save_md5 is True.
    It saves the code into the text file of 'md5_path'.
    If md5_path is None, the name of the file is the same as the input file
    with the md5 extension.

    Parameters
    ----------
        file_path: str
            Path of the file to make the MD5.
        save_md5: bool
            If save_md5 is True, it creates a file with the MD5.
        md5_path: path
            Path of the MD5 file. If md5_path is None, the name of the file is
            the same as the input file with the md5 extension.
    
    Returns
    -------
        haser: str
            MD5 code.
    """

    # Make the MD5 code
    haser = hashlib.md5()
    with open(file_path, 'rb') as open_file:
        content = open_file.read()
        haser.update(content)
    
    if save_md5:
        if md5_path:
            filename_md5 = md5_path
        else:
            # Get the same path but with md5 extension
            parts = file_path.split('.')
            filename_md5 = ""
            for part in parts[:-1]:
                filename_md5 += part
                filename_md5 += '.'
            # write the _<extension>
            filename_md5 = filename_md5[:-1] + '_'
            filename_md5 += parts[-1]

            filename_md5 += ".md5"

        # Save the md5 code into the file
        with open(filename_md5, 'w') as open_file:
            open_file.write(haser.hexdigest())

    return haser.hexdigest()



import requests
import json
from typing import Dict, List, Tuple
import urllib.parse

url = 'http://api.emso.eu'

class EMSO():
    """
    Manage the EMSO ERIC API (http://api.emso.eu)

    Parameters
    ----------
        user: str
            User or email of the EMSO ERIC API.
        password: str
            Password for the EMSO ERIC API.
        token: str
            Token for the EMSO ERIC API.
    """

    def __init__(self, user: str='', password: str='', token: str=''):

        # Creation of a requests session
        self.api = requests.Session()
        self.token = token

        if token:
            # Update requests session with the authorization header
            self.api.headers.update({'Authorization': self.token})
        elif user and password:
            # Get API token
            query = f'{url}/user/authorization?user={user}&password={password}'
            r = requests.get(query)
            if r.status_code == 201:
                self.token = r.json()['result']
                # Update requests session with the authorization header
                self.api.headers.update({'Authorization': self.token})
            else:
                raise Exception('Incorrect user or password')

    def post_user_email(self, message: Dict[str, str]) -> str:
        """
        Send an email to help@emso-eu.org

        Parameters
        ----------
            message: dict
                {'subject': 'Subject of the email',
                 'content': 'Content of the email'}
        
        Returns
        -------
            response_message: str
                Response from the API
        """
        query = f'{url}/user/email'
        message_json = json.dumps(message)

        response = self.api.post(query, payload=message_json)
        response_message = response.json()['message']
        return response_message

    def get_user_query(self, size: int=10, sort: str='desc') -> List[str]:
        """
        Get the queries of the user

        Parameters
        ----------
            size: int
                Number of results (def: 10, max:10000)
            sort: str
                Order of the values (def: 'desc', options: 'asc', 'desc')
        Returns
        -------
            query_list: List[str]
                List of the queries of the user
        """
        query = f'{url}/user/query?size={size}&sort={sort}'

        query_list = []
        r = self.api.get(query)
        if r.status_code == 201:
            query_list = r.json()['result']
        
        return query_list

    def get_info_fig(self) -> List[str]:
        """
        Get the available figures

        Returns
        -------
            fig_list: List[str]
                List of available figures
        """
        fig_list = []

        query = f'{url}/info/fig'
        r = self.api.get(query)
        if r.status_code == 200:
            fig_list = r.json()['result']

        return fig_list

    def get_info_fig_plot(self, plot: str) -> List[str]:
        """
        Get the available arguments for the input plot

        Parameters
        ----------
            plot: str
                Figure type
        
        Returns
        -------
            argument_list: List[str]
                List of available arguments
        """
        argument_list = []

        query = f'{url}/info/fig/{plot}'
        r = self.api.get(query)
        if r.status_code == 200:
            argument_list = r.json()['result']

        return argument_list

    def get_info_fig_plot_argument(
        self, plot:str, argument:str) -> Tuple[str, List[str]]:
        """
        Get argument information and available options

        Parameters
        ----------
            plot: str
                Figure type
            argument: str
                Figure argument

        Returns
        -------
            argument_info: Tuple[str, List[str]]
                (meaning of the argument, list of available options)
        """
        message = ''
        option_list = []

        query = f'{url}/info/fig/{plot}/{argument}'
        r = self.api.get(query)
        if r.status_code == 200 or r.status_code == 201:
            r_json = r.json()
            message = r_json['message']
            option_list = r_json['result']

        return (message, option_list)

    def get_info_metadata_id(self, platform_codes:List[str]=[], sites:List[str]=[]) -> List[str]:
        """
        Get the ID of the metadata archivements of the EMSO ERIC API

        Parameters
        ----------
            platform_codes: List[str]
                List of platform_code
            sites: List[str]
                List of site
        
        Returns
        -------
            metadata_ids: List[str]
                List of 'metadata_id'
        """
        metadata_ids = []

        # Make query
        query = f'{url}/info/metadata_id'
        if platform_codes or sites:
            query += '?'
            if platform_codes:
                query += f'platform_code={(",").join(platform_codes)}'
            if sites:
                if query[-1] != '?':
                    query += '&'
                query += f'site={(",").join(sites)}'

        r = self.api.get(query)
        if r.status_code == 201:
            metadata_ids = r.json()['result']
        return metadata_ids

    def get_info_parameter(self, platform_codes:List[str]=[], sites:List[str]=[]) -> List[str]:
        """
        Get available parameters of the EMSO ERIC API

        Parameters
        ----------
            platform_codes: List[str]
                List of platform_code
            sites: List[str]
                List of site
        
        Returns
        -------
            parameters: List[str]
                List of 'parameter'
        """
        parameters = []

        # Make query
        query = f'{url}/info/parameter'
        if platform_codes or sites:
            query += '?'
            if platform_codes:
                query += f'platform_code={(",").join(platform_codes)}'
            if sites:
                if query[-1] != '?':
                    query += '&'
                query += f'site={(",").join(sites)}'

        # Send the query
        r = self.api.get(query)
        if r.status_code == 201:
            parameters = r.json()['result']
        return parameters

    def get_info_platform_code(self, parameters:List[str]=[], sites:List[str]=[]) -> List[str]:
        """
        Get available platfom codes ('platform_code') of the EMSO ERIC API

        Parameters
        ----------
            parameters: List[str]
                List of 'parameter'
            sites: List[str]
                List of site
        
        Returns
        -------
            platform_codes: List[str]
                List of 'platform_code'
        """
        platform_codes = []

        # Make query
        query = f'{url}/info/platform_code'
        if parameters or sites:
            query += '?'
            if parameters:
                query += f'parameter={(",").join(parameters)}'
            if sites:
                if query[-1] != '?':
                    query += '&'
                query += f'site={(",").join(sites)}'

        # Send the query
        r = self.api.get(query)
        if r.status_code == 201:
            platform_codes = r.json()['result']

        return platform_codes

    def get_info_site(self) -> List[str]:
        """
        Get available sites ('site') of the EMSO ERIC API
        
        Returns
        -------
            sites: List[str]
                List of 'site'
        """
        sites = []

        # Make query
        query = f'{url}/info/site'

        # Send the query
        r = self.api.get(query)
        if r.status_code == 201:
            sites = r.json()['result']

        return sites

    def get_info_summary(self, fields:List[str]=[], parameters:List[str]=[],
                         platform_codes:List[str]=[], sites:List[str]=[]) -> List[dict]:
        """
        Get the most used fields of the metadatada archivements

        Parameters
        ----------
            fields: List[str]
                    List of fields to return
            parameters: List[str]
                List of 'prameter'
            platform_codes: List[str]
                List of 'platform_code'
            sites: List[str]
                List of 'site'

        Returns
        -------
            sumaries: List[dict]
                List of metadata archivements
        """
        sumaries = []

        # Make the query
        query = f'{url}/info/summary'

        if fields or parameters or platform_codes or sites:
            query += '?'
            if fields:
                query += f'field={(",").join(fields)}'
            if parameters:
                if query[-1] != '?':
                    query += '&'
                query += f'parameter={(",").join(parameters)}'
            if platform_codes:
                if query[-1] != '?':
                    query += '&'
                query += f'platform_code={(",").join(platform_codes)}'
            if sites:
                if query[-1] != '?':
                    query += '&'
                query += f'site={(",").join(sites)}'

        # Send the query
        r = self.api.get(query)
        if r.status_code == 201:
            sumaries = r.json()['result']
        return sumaries

    def get_metadata(self, fields:List[str]=[], metadata_ids:List[str]=[],
                     platform_codes:List[str]=[], sites:List[str]=[]) -> List[dict]:
        """
        Get all fields of the metadatada archivements

        Parameters
        ----------
            fields: List[str]
                    List of fields to return
            metadata_ids: List[str]
                List of 'metadata_id'
            parameters: List[str]
                List of 'prameter'
            platform_codes: List[str]
                List of 'platform_code'
            sites: List[str]
                List of 'site'

        Returns
        -------
            metadatas: List[dict]
                List of metadata archivements
        """
        metadatas = []

        # Make the query
        query = f'{url}/metadata'

        if fields or metadata_ids or platform_codes or sites:
            query += '?'
            if fields:
                query += f'field={(",").join(fields)}'
            if metadata_ids:
                if query[-1] != '?':
                    query += '&'
                query += f'metadata_id={(",").join(metadata_ids)}'
            if platform_codes:
                if query[-1] != '?':
                    query += '&'
                query += f'platform_code={(",").join(platform_codes)}'
            if sites:
                if query[-1] != '?':
                    query += '&'
                query += f'site={(",").join(sites)}'

        # Send the query
        r = self.api.get(query)
        if r.status_code == 201:
            metadatas = r.json()['result']

        return metadatas

    def get_data(self, depth_max:float=None, depth_min:float=None, depth_qcs:List[int]=[],
                 end_time:str='', metadata_ids:List[str]=[], size:int=10, sort:str='desc',
                 parameters:List[str]=[], platform_codes:List[str]=[], start_time:str='',
                 time_qcs:List[int]=[], value_qcs:List[int]=[]) -> List[dict]:
        """
        Get data from the EMSO ERIC API

        Parameters
        ----------
            depth_max: int
                Maximun depth of the measurement
            depth_min: int
                Minimum depth of the measurement
            depth_qcs: List[int]
                List of QC values accepted for the depth_qc field
            end_time: str
                Maximum date of the measurement
            metadata_ids: List[str]
                List of accepted 'metadata_id'
            size: int
                Number of values to be returned
            sort: str
                Options: 'asc' or 'desc'. Get the first or last (in time) measurements
            parameters: List[str]
                List of accepted 'parameter'
            platform_codes: List[str]
                List of accepted 'platform_code'
            start_time: str
                Minimun date of the meassurement
            time_qcs: List[int]
                List of accepted values for the field time_qc
            value_qcs: List[int]
                List of accepted values for the field of value_qc
        
        Returns
        -------
            data_list: List[dict]
                The data
        """
        data_list = []

        if size is None:
            size = 10
        if sort is None:
            sort = 'desc'

        # Make the query
        query = f'{url}/data?size={size}&sort={sort}'

        if depth_max or depth_min or depth_qcs or end_time or metadata_ids or \
            parameters or platform_codes or start_time or time_qcs or value_qcs:

            if depth_max:
                query += f'&depth_max={depth_max}'
            if depth_min:
                query += f'&depth_min={depth_min}'
            if depth_qcs:
                depth_qcs_str = map(str, depth_qcs)
                query += f'&depth_qc={(",").join(depth_qcs_str)}'
            if end_time:
                query += f'&end_time={end_time}'
            if metadata_ids:
                query += f'&metadata_id={(",").join(metadata_ids)}'
            if parameters:
                query += f'&parameter={(",").join(parameters)}'
            if platform_codes:
                query += f'&platform_code={(",").join(platform_codes)}'
            if start_time:
                query += f'&start_time={start_time}'
            if time_qcs:
                time_qcs_str = map(str, time_qcs)
                query += f'&time_qc={(",").join(time_qcs_str)}'
            if value_qcs:
                value_qcs_str = map(str, value_qcs)
                query += f'&value_qc={(",").join(value_qcs_str)}'

        # Send the query
        r = self.api.get(query)
        if r.status_code == 201:
            data_list = r.json()['result']

        return data_list

    def get_fig_data_interval(self, interval:str='D', depth_max:int=None, depth_min:int=None,
                              depth_qcs:List[int]=[], end_time:str='', metadata_ids:List[str]=[],
                              parameters:List[str]=[], platform_codes:List[str]=[],
                              rangeslider:bool=False, start_time:str='', time_qcs:List[int]=[],
                              title:str='', value_qcs:List[int]=[]) -> dict:
        """
        Get the plotly figure 'Data Interval' from the EMSO ERIC API

        Parameters
        ----------
            interval: str
                Period to find a data value
            depth_max: int
                Maximun depth of the measurement
            depth_min: int
                Minimum depth of the measurement
            depth_qcs: List[int]
                List of QC values accepted for the depth_qc field
            end_time: str
                Maximum date of the measurement
            metadata_ids: List[str]
                List of accepted 'metadata_id'
            parameters: List[str]
                List of accepted 'parameter'
            platform_codes: List[str]
                List of accepted 'platform_code'
            rangeslider: bool
                Enable a range slider on to bottom of the graph
            start_time: str
                Minimun date of the meassurement
            time_qcs: List[int]
                List of accepted values for the field time_qc
            title: str
                Title of the figure
            value_qcs: List[int]
                List of accepted values for the field of value_qc
        
        Returns
        -------
            fig: dict
                Plotly figure
        """
        # Add the number
        if len(interval) < 2:
            interval = '1' + interval

        interval_str = interval.replace(
            'S', 's').replace('T', 'm').replace('H', 'h').replace('D', 'd')

        fig = {}

        if rangeslider:
            rangeslider_str = 'true'
        else:
            rangeslider_str = 'false'

        # Make the query
        query = f'{url}/fig/data_interval/{interval_str}?rangeslider={rangeslider_str}'

        print(query)

        if depth_max or depth_min or depth_qcs or end_time or metadata_ids or \
            parameters or platform_codes or start_time or time_qcs or \
            value_qcs or title or rangeslider:

            if depth_max:
                query += f'&depth_max={depth_max}'
            if depth_min:
                query += f'&depth_min={depth_min}'
            if depth_qcs:
                depth_qcs_str = map(str, depth_qcs)
                query += f'&depth_qc={(",").join(depth_qcs_str)}'
            if end_time:
                query += f'&end_time={end_time}'
            if metadata_ids:
                query += f'&metadata_id={(",").join(metadata_ids)}'
            if parameters:
                query += f'&parameter={(",").join(parameters)}'
            if platform_codes:
                query += f'&platform_code={(",").join(platform_codes)}'
            if start_time:
                query += f'&start_time={start_time}'
            if time_qcs:
                time_qcs_str = map(str, time_qcs)
                query += f'&time_qc={(",").join(time_qcs_str)}'
            if title:
                query += f'&title={title}'
            if value_qcs:
                value_qcs_str = map(str, value_qcs)
                query += f'&value_qc={(",").join(value_qcs_str)}'

        # Send the query
        r = self.api.get(query)
        if r.status_code == 201:
            fig = r.json()['result']

        return fig

    def get_fig_line(self, depth_max:int=None, depth_min:int=None, depth_qcs:List[int]=[],
                     end_time:str='', metadata_ids:List[str]=[], parameters:List[str]=[],
                     platform_codes:List[str]=[], size:int=10, start_time:str='', sort:str='desc',
                     time_qcs:List[int]=[], title:str='', value_qcs:List[int]=[], x='time',
                     y='value') -> dict:
        """
        Get the plotly figure 'line' from the EMSO ERIC API

        Parameters
        ----------
            depth_max: int
                Maximun depth of the measurement
            depth_min: int
                Minimum depth of the measurement
            depth_qcs: List[int]
                List of QC values accepted for the depth_qc field
            end_time: str
                Maximum date of the measurement
            metadata_ids: List[str]
                List of accepted 'metadata_id'
            parameters: List[str]
                List of accepted 'parameter'
            platform_codes: List[str]
                List of accepted 'platform_code'
            size: int
                Number of values to make the graph
            start_time: str
                Minimun date of the meassurement
            sort: str
                Options: 'asc' or 'desc'. Get the first or last (in time) measurements
            time_qcs: List[int]
                List of accepted values for the field time_qc
            title: str
                Title of the figure
            value_qcs: List[int]
                List of accepted values for the field of value_qc
            x: str
                Field for the x-axis
            y: str
                Field for the y-axis
        Returns
        -------
            fig: dict
                Plotly figure
        """
        fig = {}

        if size is None:
            size = 10
        if sort is None:
            sort = 'desc'
        if x is None:
            x = 'time'
        if y is None:
            y = 'value'

        # Make the query
        query = f'{url}/fig/line?size={size}&sort={sort}&x={x}&y={y}'

        if depth_max or depth_min or depth_qcs or end_time or metadata_ids or \
            parameters or platform_codes or start_time or time_qcs or value_qcs:

            if depth_max:
                query += f'&depth_max={depth_max}'
            if depth_min:
                query += f'&depth_min={depth_min}'
            if depth_qcs:
                depth_qcs_str = map(str, depth_qcs)
                query += f'&depth_qc={(",").join(depth_qcs_str)}'
            if end_time:
                query += f'&end_time={end_time}'
            if metadata_ids:
                query += f'&metadata_id={(",").join(metadata_ids)}'
            if parameters:
                query += f'&parameter={(",").join(parameters)}'
            if platform_codes:
                query += f'&platform_code={(",").join(platform_codes)}'
            if start_time:
                query += f'&start_time={start_time}'
            if time_qcs:
                time_qcs_str = map(str, time_qcs)
                query += f'&time_qc={(",").join(time_qcs_str)}'
            if title:
                query += f'&title={title}'
            if value_qcs:
                value_qcs_str = map(str, value_qcs)
                query += f'&value_qc={(",").join(value_qcs_str)}'

        # Send the query
        r = self.api.get(query)
        if r.status_code == 201:
            fig = r.json()['result']

        return fig

    def get_fig_map(self, parameters:List[str]=[], platform_codes:List[str]=[],
                    sites:List[str]=[]) -> dict:
        """
        Get the plotly figure 'Map' from the EMSO ERIC API

        Parameters
        ----------
            parameters: List[str]
                List of accepted 'parameter'
            platform_codes: List[str]
                List of accepted 'platform_code'
            site: List[str]
                List of accepted 'site'
        Returns
        -------
            fig: dict
                Plotly figure
        """
        fig = {}

        # Make the query
        query = f'{url}/fig/map'

        if parameters or platform_codes or sites:
            query += '?'
            if parameters:
                query += f'parameter={(",").join(parameters)}'
            if platform_codes:
                if query[-1] != '?':
                    query += '&'
                query += f'platform_code={(",").join(platform_codes)}'
            if sites:
                if query[-1] != '?':
                    query += '&'
                query += f'site={(",").join(sites)}'

        # Send the query
        r = self.api.get(query)
        if r.status_code == 201:
            fig = r.json()['result']

        return fig



""" Implementation of WaterFrame.min(parameter_name) """
import numpy as np

def min(self, parameter_min):
    """
    It returns the minimum value of a parameter and the value's indexes.

    Parameters
    ----------
        parameter_min: str
            Name of the parameter.

    Returns
    -------
        min_dict: dict
            Dictionary with the following format:
            {
                '<name of index 1>': <value of index 1>,
                '<name of index n>': <value of index n>,
                'name of parameter': < min value of parameter>
            }
            If min_dict is None, all the values of the parameter are NaN.
    """

    df = self.data[parameter_min]
    df = df.reset_index()

    try:
        min_dict = df.loc[df[parameter_min] == df[parameter_min].min(skipna=True)].to_dict('record')[0]
    except IndexError:
        min_dict = None

    return min_dict



""" Implementation of WaterFrame.max(parameter_max) """


def max(self, parameter_max):
    """
    It returns the maximum value of a parameter and the value's indexes.

    Parameters
    ----------
        parameter_max: str
            Name of the parameter.

    Returns
    -------
        max_dict: dict
            Dictionary with the following format:
            {
                '<name of index 1>': <value of index 1>,
                '<name of index n>': <value of index n>,
                'name of parameter': < maximum value of parameter>
            }
            If max_dict is None, all the values of the parameter are NaN.
    """

    df = self.data[parameter_max]
    df = df.reset_index()

    try:
        max_dict = df.loc[
            df[parameter_max] == df[parameter_max].max(skipna=True)].to_dict('record')[0]
    except IndexError:
        max_dict = None

    return max_dict



from .iplot_location import iplot_location
from .iplot_timeseries import iplot_timeseries
from .iplot import iplot
from .iplot_scatter import iplot_scatter
from .iplot_line import iplot_line
from .iplot_data_intervals import iplot_data_intervals



""" Implementation of WaterFrame.resample(rule, method='mean') """
import pandas as pd

def resample(self, rule, method='mean', inplace=True):
    """
    Convenience method for frequency conversion and sampling of time series of the WaterFrame
    object. Warning: if WaterFrame.data contains MultiIndex, those indexes will disappear, obtaining a single 'TIME'
    index.

    Parameters
    ----------
        rule: str
            The offset string or object representing target conversion.
            You can find all of the resample options here:
            http://pandas.pydata.org/pandas-docs/stable/timeseries.html#offset-aliases
        method: "mean", "max", "min", optional, (method = "mean")
            Save the new value with the mean(), max() or min() function.
        inplace: bool
            If True, resample in place and return 'True', If False, return a new WaterFrame.

    Returns
    -------
        new_wf: WaterFrame
    """
    if isinstance(self.data.index,
        pd.MultiIndex) and 'DEPTH' in self.data.index.names:
    
        # pd.Grouper allows you to specify a "groupby instruction for a target
        # object".
        # Ref: https://stackoverflow.com/questions/15799162/resampling-within-a-pandas-multiindex
        level_values = self.data.index.get_level_values

        if method == "mean":
            data = (
                self.data.groupby(
                    [level_values(0)] +  # 0 is 'DEPTH'
                    [pd.Grouper(freq=rule, level='TIME')]).mean())
        elif method == "max":
            data = (
                self.data.groupby(
                    [level_values(0)] +  # 0 is 'DEPTH'
                    [pd.Grouper(freq=rule, level='TIME')]).max())
        elif method == "min":
            data = (
                self.data.groupby(
                    [level_values(0)] +  # 0 is 'DEPTH'
                    [pd.Grouper(freq=rule, level='TIME')]).min())
    else:
        # Legacy
        if method == "mean":
            data = self.data.resample(rule, level='TIME').mean()
        elif method == "max":
            data = self.data.resample(rule, level='TIME').max()
        elif method == "min":
            data = self.data.resample(rule, level='TIME').min()

    # Change "_QC" values to 0
    for key in self.data.keys():
        if "_QC" in key:
            data[key] = 0

    if inplace:
        self.data = data
        return True
    else:
        new_wf = self.copy()
        new_wf.data = data
        return new_wf



""" Implementation of wf.qc_replace() """

def qc_replace(self, parameters=None, to_replace=0, value=1, start=0, inplace=True):
    """
    Replace the values of QC from the input parameters.

    Parameters
    ----------
        parameters: None, str or list of str.
            List of parameters to change the values of QC. Ex: ['TEMP', 'PSAL'].
        to_replace: int
            QC value to replace.
        value: int
            Value to replace any values matching to_replace with.
        inplace: bool
            If inplace, makes changes inplace and returns True.
            Otherwhise, returns a new WaterFrame.
    
    Returns
    -------
        new_wf: WaterFrame
    """
    def change_signals(signals):
         # Change flags
        result = []
        for i, signal in enumerate(signals):
            if i < start:
                result.append(signal)
            else:
                if signal == to_replace:
                    result.append(value)
                else:
                    result.append(signal)

        return result

    if parameters is None:
        parameters = self.parameters
    elif isinstance(parameters, str):
        parameters = [parameters]

    data = self.data.copy()

    for parameter in parameters:
        
        # New
        df = data[[parameter, f'{parameter}_QC']].reset_index()
        df.set_index('TIME', inplace=True)

        try:
            for depth, df_depth in df.groupby('DEPTH'):
                df_depth.sort_index(inplace=True)

                signals = df_depth[f'{parameter}_QC'].values

                # Change flags
                result = change_signals(signals)

                data.loc[(depth,), f'{parameter}_QC'] = result
        except KeyError:
            # No Depth
            signals = df[f'{parameter}_QC'].values

            # Change flags
            result = change_signals(signals)
            
            data[f'{parameter}_QC'] = result

    if inplace:
        self.data = data
        return True
    else:
        new_wf = self.copy()
        new_wf.data = data
        return new_wf


from .copy import copy
from .corr import corr
from .min import min
from .max import max
from .use_only import use_only
from .rename import rename
from .max_diff import max_diff
from .time_intervals import time_intervals
from .resample import resample
from .slice_time import slice_time
from .info_metadata import info_metadata
from .info_vocabulary import info_vocabulary
from .drop import drop
from .reduce_memory import reduce_memory



import pandas as pd
import plotly.express as px
import plotly.graph_objects as go


def iplot_line(self, y, x='TIME', marginal_x=None, marginal_y='histogram', color='DEPTH',
               range_y='auto', line_shape='lineard', rangeslider_visible=True, **kwds):
    """
    It uses plotly.express.line.
    Each data point is represented as a marker point, whose location is given by the x and y columns
    of self.data.

    Parameters
    ----------
        y: str
            Y axes, column or index of data.
        x: str
            X axes, column or index of data.
        color: str
            Name of a column or index of data. Values from this column are used to assign color to
            marks. If color = 'auto', color = QC column of y.
        range_y: list
            [min value, max value] of y axes. If range_y = 'auto', range is generated between the
            mina nd max values of y axes +- 5%.
        line_shape: str
            Line options: 'linear' 'spline', 'vhv', 'hvh', 'vh', 'hv'
        rangeslider_visible: bool
            Show a time range slide on the bottom x axes.
        **kwds: keywords
            plotly express scatter keywords.

    Returns
    -------
        fig: plotly.graph_objects.Figure
    """
    df = pd.DataFrame()
    _df = self.data.reset_index()

    df[y] = _df[y]
    df[f'{y}_QC'] = _df[f'{y}_QC']
    df[x] = _df[x]

    if 'DEPTH' in _df.keys():
        df['DEPTH'] = _df['DEPTH']

    # Range Y calculation
    y_min = min(df[y].values)
    y_max = max(df[y].values)
    y_percent = (y_max - y_min) / 100

    if range_y == 'auto':
        range_y = [y_min - 5* y_percent, y_max + 5* y_percent]

    # Save memory
    del _df

    # Dropna
    df.dropna(inplace=True)

    if 'DEPTH' in df.keys() and 'TIME' in df.keys():
        # Sort by TIME AND DEPTH
        df.sort_values(['DEPTH', 'TIME'], inplace=True)
    elif 'TIME' in df.keys():
        df.sort_values(['TIME'], inplace=True)

    if color == 'DEPTH':
        df[color] = df[color].astype('str')
    elif color:
        df[color] = df[color]

    # # Set index TIME
    # df.set_index('TIME', inplace=True)
    # df.sort_index(inplace=True)
    # df.reset_index(inplace=True)

    fig = px.line(df, x=x, y=y, color=color,
                  range_y=range_y,
                  line_shape=line_shape,
                  labels={
                      y: self.vocabulary[y].get('units', y)},
                  **kwds)

    fig.update_xaxes(rangeslider_visible=rangeslider_visible)
    fig.update_layout(margin=dict(l=30, r=0, t=30, b=0))

    if 'QC' in color:
        fig.for_each_trace(
            lambda trace: trace.update(
                visible='legendonly',
                mode='markers',
                marker_color='red') if trace.name == 'Bad data' else (),
        )
        fig.for_each_trace(
            lambda trace: trace.update(
                mode='lines+markers',
                marker_color='blue',
                line_color='blue') if trace.name == 'Good data' else (),
        )
    elif 'DEPTH' in color:
        fig.for_each_trace(
            lambda trace: trace.update(mode='lines+markers'))

    return fig



""" Implementation of a function to get a WaterFrame object from a JSON. """
import json
import pandas as pd


def read_json(self, json_string):
    """
    It gets a WaterFrame object from a JSON.

    Parameters
    ----------
        json_string: str
            String that contains a JSON.

    Returns
    -------
        done: bool
            True if the operation is successful.
    """

    done = False
    try:
        big_dict = json.loads(json_string)

        keys = big_dict.keys()
        if "metadata" in keys:
            self.metadata = big_dict["metadata"].copy()
        if "vocabulary" in keys:
            self.meaning = big_dict["vocabulary"].copy()
        if "data" in keys:
            self.data = pd.read_json(big_dict["data"])
        else:
            self.data = pd.read_json(big_dict)

        done = True
    except ValueError:
        pass
    return done



""" Implementation of WaterFrame.plot_timeseries(keys=None, rolling=None, ax=None,
average_time=None, secondary_y=False, color=None)"""


def plot_timeseries(self, parameters_to_plot=None, qc_flags=None, rolling_window=None, ax=None,
                    average_time=None, secondary_y=None, color=None):
    """
    Plot the input parameters with time on X and the parameters on Y. It calculates the
    standar deviation of a rolling window and plot it.

    Parameters
    ----------
        parameters_to_plot: list of str, str, optional (parameters_to_plot = None)
            Parameters of the WaterFrame to plot. If parameters_to_plot is None, all parameters
            will be ploted.
        qc_flags: list of int, optional (qc_flags = None)
            QC flags of the parameters to plot. If qc_flags in None, all QC flags will be used.
        rolling_window: int, optional (rolling_window = None)
            Size of the moving window. It is the number of observations
            used for calculating the statistic.
        ax: matplotlib.axes object, optional (ax = None)
            It is used to add the plot to an input axes object.
        average_time: str, optional (average_time = None)
            It calculates an average value of a time interval. You can find
            all of the resample options here:
            http://pandas.pydata.org/pandas-docs/stable/timeseries.html#offset-aliases
        secondary_y: bool, optional (secondary_y = False)
            Plot on the secondary y-axis.
        color: str or list of str, optional (color = None)
            Any matplotlib color. It will be applied to the traces.
    Returns
    -------
        ax: matplotlib.AxesSubplot
            Axes of the plot.
    """
    def make_plot(df_in, parameter_name, ax_in, color_in, rolling_in):
        """
        It makes the graph.

        Parameters
        ----------
            df_in: pandas.DataFrame
                It contains the info to plot.
            parameter_name: str
                Name of the parameter to plot.
            ax_in: matplotlib.axes
                Axes to add the plot.
            color_in: str
                Color of the line.
            rolling_in: int
                Size of the rolling window to calculate the mean and standar deviation.

        Returns
        -------
            ax_out: matplotlib.axes
                Axes with the plot.
        """
        # Calculation of the std and the mean
        roll = df_in[parameter_name].rolling(rolling_in, center=True)
        m = roll.agg(['mean', 'std'])
        # rename 'mean' column
        m.rename(columns={'mean': parameter_name}, inplace=True)
        m.dropna(inplace=True)
        ax_out = m[parameter_name].plot(ax=ax_in, secondary_y=secondary_y, legend=True,
                                        color=color_in)
        ax_out.fill_between(m.index, m[parameter_name] - m['std'], m[parameter_name] + m['std'],
                            alpha=.25, color=color_in)
        ax_out.legend([f"{parameter_name} at depth: {df_in['DEPTH'].mean()}"])
        # Write axes
        try:
            ax_out.set_ylabel(self.vocabulary[parameter_name]['units'])
        except KeyError:
            # No units
            pass

        return ax_out

    if parameters_to_plot is None:
        parameters_to_plot = self.parameters
    elif isinstance(parameters_to_plot, str):
        parameters_to_plot = [parameters_to_plot]

    # Extract data
    # Dropna is necessary?
    df = self.data[parameters_to_plot].dropna().reset_index()
    # df = self.data[parameters_to_plot].dropna().reset_index().set_index('TIME')
    df = df.groupby(['DEPTH', 'TIME'])[parameters_to_plot].mean()
    # print(df2)
    df.reset_index(inplace=True)
    df.set_index('TIME', inplace=True)

    for _, df_depth in df.groupby('DEPTH'):
        df_depth.index.rename("Date", inplace=True)
        df_depth.sort_index(inplace=True)

        # Resample data
        if average_time is None:
            pass
        else:
            df_depth = df_depth.resample(average_time).mean()

        # Calculation of the rolling value
        if rolling_window is None:
            if df_depth.size <= 100:
                rolling_window = 1
            elif df_depth.size <= 1000:
                rolling_window = df_depth.size//10
            elif df_depth.size <= 10000:
                rolling_window = df_depth.size // 100
            else:
                rolling_window = df_depth.size // 1000

        if color is None:
            for parameter_to_plot in parameters_to_plot:
                ax = make_plot(df_depth, parameter_to_plot, ax, color, rolling_window)
        else:
            for parameter_to_plot, color_to_plot in zip(parameters_to_plot, color):
                ax = make_plot(df_depth, parameter_to_plot, ax, color_to_plot, rolling_window)

    return ax



""" Implementation of WaterFrame.use_only(*parameters*) """


def use_only(self, parameters_to_use, inplace=True):
    """
    It deletes all parameters except the input parameters.

    Parameters
    ----------
        parameters_to_use: list of str
            List of parameter names.
        inplace: bool
            If True, do operation inplace and return True.

    Returns
    -------
        new_wf: WaterFrame
            Result of the WaterFrame after the method application.
    """

    # Force to have a list in parameters
    if isinstance(parameters_to_use, str):
        parameters_to_use = [parameters_to_use]

    # Check if all parameters_to_use exist
    parameters_qc = []
    for parameter_to_use in parameters_to_use:
        if parameter_to_use not in self.parameters:
            raise KeyError(f"{parameter_to_use} is not a parameter of the WaterFrame.")
        else:
            parameters_qc.append(f"{parameter_to_use}_QC")
    parameters_to_use += parameters_qc

    if inplace:
        self.data = self.data[parameters_to_use]
        return True
    else:
        # Drop parameters and return a new WaterFrame
        new_wf = self.copy()
        new_wf.data = self.data[parameters_to_use]
        return new_wf



""" Main implementation of the class WaterFrame """
import sys
from pandas import DataFrame


class WaterFrame:

    from .analysis import (
        min, max, copy, use_only, rename, corr, max_diff, time_intervals, resample, slice_time,
        info_metadata, info_vocabulary, drop, reduce_memory)
    from .output import to_nc, to_pkl, to_json, to_es, metadata_to_html, to_csv
    from .plot import plot_timeseries, plot_timebar, plot_hist, plot
    from .qc import qc_flat_test, qc_range_test, qc_spike_test, qc_replace, qc_syntax_test
    from .iplot import (
        iplot_location, iplot_timeseries, iplot, iplot_scatter, iplot_line,
        iplot_data_intervals)

    def __init__(self, df=None, metadata=None, vocabulary=None):
        """ Constructor """
        if df:
            self.data = df
        else:
            self.data = DataFrame()

        if metadata:
            self.metadata = metadata
        else:
            self.metadata = dict()

        if vocabulary:
            self.vocabulary = vocabulary
        else:
            self.vocabulary = dict()

    def __repr__(self):
        """ Implementation of repr """

        # Memory use message
        memory_bytes = self.memory_usage
        units = "Bytes"
        if memory_bytes > 1000000000:
            memory_bytes /= 1000000000
            units = "GBytes"
        elif memory_bytes > 1000000:
            memory_bytes /= 1000000
            units = "MBytes"
        elif memory_bytes > 1000:
            memory_bytes /= 1000
            units = "KBytes"
        size_message = f"Memory usage: {memory_bytes:.3f} {units}"

        # Parameters message
        parameters_message = "Parameters:"
        for parameter in self.parameters:
            try:
                parameters_message += f"\n  - {parameter}: " + \
                    f"{self.vocabulary[parameter]['long_name']}" + \
                    f" ({self.vocabulary[parameter]['units']})"
            except KeyError:
                parameters_message += \
                    "\n  - {}: Parameter without meaning".format(parameter)

            # Min, max and mean info
            min_dict = self.min(parameter)
            max_dict = self.max(parameter)
            value_mean = self.data[parameter].mean()
            if min_dict is not None and max_dict is not None:
                # min value string
                parameters_message += f"\n    - Min value: {min_dict[parameter]}"
                for key, value in min_dict.items():
                    if key == parameter:
                        continue
                    else:
                        parameters_message += f"\n      - {key}: {value}"
                # max value string
                parameters_message += f"\n    - Max value: {max_dict[parameter]}"
                for key, value in max_dict.items():
                    if key == parameter:
                        continue
                    else:
                        parameters_message += f"\n      - {key}: {value}"
                # mean value string
                parameters_message += f"\n    - Mean value: {value_mean}"
            else:
                parameters_message += "\n    - Parameter without values."

        message = size_message + "\n" + parameters_message

        return message

    def _repr_html_(self, *args, **kwargs):
        data_html = self.data._repr_html_()

        metadata_list = ''
        for key, value in self.metadata.items():
            metadata_list += f'<li><b>{key}</b>: {value}</li>'

        dictionary_html = f'<div><p>Metadata:</p><ul>{metadata_list}</ul><p>Data:</p></div>'
        return dictionary_html + data_html

    @property
    def parameters(self):
        """
        Get the keys of data with "QC" columns.
        """
        return_keys = []
        keys = self.data.keys()
        for key in keys:
            try:
                if "_QC" in key:
                    continue
                for key_qc in keys:
                    if "_QC" not in key_qc:
                        continue
                    if key_qc == f"{key}_QC":
                        return_keys.append(key)
            except TypeError:
                # TypeError: argument of type 'Timestamp' is not iterable
                continue
        return return_keys

    @property
    def memory_usage(self):
        """
        It returns the memory usage of the WaterFrame in bytes.
        """
        size = sys.getsizeof(self.data) + sys.getsizeof(self.metadata)
        return size

    @property
    def empty(self):
        """
        It return True if the data is empty.

        Returns
        -------
            empty: bool
        """
        empty = self.data.empty
        return empty



""" Implementation of WaterFrame.corr(parameter1, parameter2, method='pearson', min_periods=1) """


def corr(self, parameter1, parameter2, method='pearson', min_periods=1):
    """
    Compute pairwise correlation of data columns of parameter1 and parameter2, excluding NA/null
    values.

    Parameters
    ----------
        parameter1: str
            Key name of the column 1 to correlate.
        parameter2: str
            Key name of the column 2 to correlate.
        method: {‘pearson’, ‘kendall’, ‘spearman’}
            pearson : standard correlation coefficient
            kendall : Kendall Tau correlation coefficient
            spearman : Spearman rank correlation
        min_periods : int, optional
            Minimum number of observations required per pair of columns to
            have a valid result. Currently only available for pearson and
            spearman correlation
    Returns
    -------
        correlation_number: float
    """
    correlation_number = self.data[
        parameter1].corr(self.data[parameter2], method=method, min_periods=min_periods)
    return correlation_number



import flask
import os


def metadata_to_html(self, filename=None, logo_link=None, footer_text=''):
    """
    Make a html file with the metadata information.

    Parameters
    ----------
        filename: str
            Filename and path of the html file. If filename is None, the file will be saved in the
            root folder with the name of the metadata['id'].
        logo_link: str
            URL with the logo to be placed on the medatada file.
        footer_text: str
            Text to be placced on the footer of the html file.
    
    Returns
    -------
        filename: str
            Name of the html file.
    """
    # Find the path to the html templates directory
    dirname = os.path.dirname(__file__)
    template_folder = os.path.join(dirname, 'html_templates')

    # Look for the filename, that is equal to self.metadata['id']
    # If self.metadata['id'] does not exist, filename is 'metadata.html'
    if filename is None:
        filename = self.metadata.get('id', 'metadata') + '.html'

    app = flask.Flask('my metadata', template_folder=template_folder)

    with app.app_context():
        rendered = flask.render_template(
            'metadata.html',
            logo_link=logo_link,
            metadata_dict=self.metadata,
            footer_text=footer_text
        )

    with open(filename, 'w') as handle:
        handle.write(rendered)
    
    return filename



""" Implementation of mooda.iplot_location() """


def iplot_location(list_wf):
    """
    It creates a Plotly Figure with a map and a spot of the measurement location of the input
    WaterFrames.

    Parameters
    ----------
        list_wf: List of WaterFrames
            List of WaterFrames to be concatenated.

    Returns
    -------
        figure: dict
            Dictionary of Plotly figure
    """

    lat = [float(wf.metadata.get('geospatial_lat_min')) for wf in list_wf]
    lon = [float(wf.metadata.get('geospatial_lon_min')) for wf in list_wf]
    text = [wf.metadata.get('platform_code') for wf in list_wf]
    # Data creation
    data = [
        dict(
            type="scattergeo",
            lon=lon,
            lat=lat,
            mode='markers',
            marker=dict(
                size=16,
            ),
            text=text
        )
    ]

    # Configuration of the zoom
    if lat is not None and lon is not None:
        lataxis = dict(range=[min(lat)-9, max(lat)+9])
        lonaxis = dict(range=[min(lon)-16, max(lon)+16])
    else:
        lataxis = None
        lonaxis = None

    layout = dict(
        geo=dict(
            lakecolor="rgb(255, 255, 255)",
            resolution=50,
            showcoastlines=True,
            showland=True,
            landcolor="rgb(229, 229, 229)",
            countrycolor="rgb(255, 255, 255)",
            coastlinecolor="rgb(255, 255, 255)",
            lataxis=lataxis,
            lonaxis=lonaxis
        ),
        margin=dict(l=10, r=10, t=0, b=0),
    )

    figure = dict(data=data, layout=layout)
    return figure



""" Implementation of md.read_nc_moist() """
from . import read_nc

def read_nc_moist(path, resample_rule=False):
    """
    Open a NetCDF file from MOIST.
    This method only works with the files generated from CTDs.

    Parameters
    ----------
        path: str
            Path of the NetCDF file.
        resample_rule: str
            Time resample method. Use resample_rule = False to do not resample.
            See DateOffset objects:
            https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html

    Returns
    -------
        wf: WaterFrame
    """

    def drop(wf_in):
        wf_in.data.reset_index(inplace=True)
        del wf_in.data['LATITUDE']
        del wf_in.data['LONGITUDE']
        for key in wf_in.data.keys():
            if 'SDN_' in key or 'SEADATANET_QC' in key or 'INST' in key:
                del wf_in.data[key]
        del wf_in.data['crs']
        del wf_in.data['MAXT']

        del wf_in.vocabulary['LATITUDE']
        del wf_in.vocabulary['LONGITUDE']
        for key in list(wf_in.vocabulary.keys()):
            if 'SDN_' in key or 'SEADATANET_QC' in key or 'INST' in key:
                del wf_in.vocabulary[key]
        del wf_in.vocabulary['crs']

        return wf_in

    def resample(wf_in, resample_rule_in):
        wf_in.data.set_index('TIME', inplace=True)
        wf_in.data = wf_in.data.resample(resample_rule_in).mean()
        return wf_in

    def set_index(wf_in):
        wf_in.data.reset_index(inplace=True)
        wf_in.data.set_index(['DEPTH', 'TIME'], inplace=True)
        return wf_in

    def rename(wf_in):

        wf_in.vocabulary['DEPTH']['ancillary_variables'] = 'DEPTH_QC'
        wf_in.vocabulary['TIME']['ancillary_variables'] = 'TIME_QC'

        wf_in.vocabulary['CNDC'] = wf_in.vocabulary.pop('conductivity')
        wf_in.vocabulary['CNDC']['ancillary_variables'] = 'CNDC_QC'

        wf_in.vocabulary['TEMP'] = wf_in.vocabulary.pop('temperature')
        wf_in.vocabulary['TEMP']['ancillary_variables'] = 'TEMP_QC'

        wf_in.vocabulary['PRES'] = wf_in.vocabulary.pop('pressure')
        wf_in.vocabulary['PRES']['ancillary_variables'] = 'PRES_QC'

        wf_in.data.rename(
            columns={
                'conductivity': 'CNDC',
                'temperature': 'TEMP',
                'pressure': 'PRES'
            }, inplace=True)
        
        wf_in.data['PRES_QC'] = 0
        wf_in.data['TEMP_QC'] = 0
        wf_in.data['CNDC_QC'] = 0
        wf_in.data['TIME_QC'] = 0
        wf_in.data['DEPTH_QC'] = 0

        return wf_in

    wf = read_nc(path)
    wf = drop(wf)
    if resample_rule:
        wf = resample(wf, resample_rule)
    wf = set_index(wf)
    wf = rename(wf)

    return wf



"""
Implementation of WaterFrame.plot_hist(parameter=None, mean_line=False, **kwds)
"""

def plot_hist(self, parameters=None, mean_line=False, **kwds):
    """
    Make a histogram of the WaterFrame's.
    A histogram is a representation of the distribution of data.
    This function calls pandas.DataFrame.hist(), on each parameter of the
    WaterFrame, resulting in one histogram per parameter.

    Parameters
    ----------
        parameters: str, list of str, optional (parameters=None)
            keys of self.data to plot. If parameters=None, it will plot all
            parameters.
        mean_line: bool, optional (mean_line=False)
                It draws a line representing the average of the values.
        **kwds:
            All other plotting keyword arguments to be passed to
            DataFrame.hist().
            https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.hist.html
    Returns
    -------
        ax: matplotlib.AxesSubplot
            Axes of the plot.
    """

    if parameters is None:
        parameters = list(self.parameters)

    if isinstance(parameters, str):
        parameters = [parameters]

    axes = self.data.hist(column=parameters, **kwds)
    parameter_counter = 0
    try:
        for ax in axes:
            ax.set_xlabel("Values")
            ax.set_ylabel("Frequency")
            if mean_line is True:
                if parameter_counter < len(parameters):
                    x_mean = self.mean(parameters[parameter_counter])
                    ax.axvline(x_mean, color='k',
                                linestyle='dashed',
                                linewidth=1)

                    parameter_counter += 1
    except AttributeError:

        # Creation of the mean line
        parameter_counter = 0
        for irow in range(len(axes)):
            for icol in range(len(axes[irow])):
                if parameter_counter < len(parameters):
                    axes[irow, icol].set_xlabel("Values")
                    axes[irow, icol].set_ylabel("Frequency")
                    if mean_line is True:
                        x_mean = self.data[axes[irow, icol].get_title()].mean()
                        axes[irow, icol].axvline(x_mean, color='k',
                                                    linestyle='dashed',
                                                    linewidth=1)
                    parameter_counter += 1

    return axes



""" Implementation of WaterFrame.slice_time(start=None, end=None) """
import pandas as pd
import datetime


def slice_time(self, start=None, end=None, inplace=True):
    """
    Delete data outside the time interval.

    Parameters
    ----------
        start: str, timestamp, optional (start=None)
            Start time interval with format 'YYYYMMDDhhmmss' or timestamp.
        end: str, timestamp, optional (end=None)
            End time interval with format 'YYYYMMDDhhmmss' or timestamp.
        inplace: bool
            If inplace is True, changes will be applied on self.data and returns True.
            Otherwhise, it returs the new WaterFrame

    Returns
    -------
        new_wf: WaterFrame
    """
    start_time = None
    end_time = None
    if start:
        start_time = datetime.datetime.strptime(start, '%Y%m%d%H%M%S')
    if end:
        end_time = datetime.datetime.strptime(end, '%Y%m%d%H%M%S')
    
    # data = self.data.sort_index(level='TIME')
    data = self.data.sort_index()

    idx = pd.IndexSlice
    data = data.loc[idx[:, :, start_time:end_time], :]

    if inplace:
        self.data = data
        return True
    else:
        new_wf = self.copy()
        new_wf.data = data
        return new_wf



""" Implementation of mooda.widget_qc()"""
from typing import List
import ipywidgets as widgets
from IPython.display import clear_output
import plotly.offline as pyo


def widget_qc(wf, parameter: str, range_test: List[float]=[-1000, 1000],
              spike_window: int=100, spike_threshold: float=3.5,
              spike_influence: float=0.5):
    """
    It makes a Data QC Widget for Jupyter Notebook

    Parameters
    ----------
        wf: WaterFrame
        parameter: str
            Parameter
        range_test: List[float]
            Limits for the range test [min value, max value]
        spike_window: int
            Window for the spike test
        spike_threshold: float
            Float for the spike test
        spike_influence: float
            Influence of the bad data on the spike test

    Returns
    -------
        main_box: ipwidgets.VBox
            Jupyter notebook widget 
    """
    pyo.init_notebook_mode()

    def show_result(wf, parameter_in, chart_title=''):
        # Change name of flags
        wf2 = wf.copy()
        qc_labels = {0: 'No QC', 1: 'Good data', 4: 'Bad data'}
        wf2.data[f'{parameter_in}_QC'].replace(qc_labels, inplace=True)

        fig = wf2.iplot_line(
            parameter_in,
            color=f'{parameter_in}_QC',
            marginal_y=None,
            line_shape='linear',
            rangeslider_visible=False,
            line_dash_sequence=['dot', 'dot'],
            title=chart_title)
        fig.show()

    # Reset flags
    reset_label = widgets.Label('Reset flags:')
    reset_checkbox = widgets.Checkbox(description='Do it!')

    # Flat test
    flat_label = widgets.Label('Flat test:')
    window_flat = widgets.IntText(value=2,
                                    description='Window:',
                                    disabled=False,
                                    color='black')
    flat_checkbox = widgets.Checkbox(description='Do it!')

    # Range test
    range_label = widgets.Label('Range test:')
    range_checkbox = widgets.Checkbox(description='Do it!')
    limits = widgets.FloatRangeSlider(value=[range_test[0], range_test[1]],
                                        min=range_test[0],
                                        max=range_test[1],
                                        step=0.1,
                                        description='Limits:',
                                        disabled=False,
                                        continuous_update=False,
                                        orientation='horizontal',
                                        readout=True,
                                        readout_format='.1f')

    # Spike test
    spike_label = widgets.Label('Spike test:')
    spike_checkbox = widgets.Checkbox(description='Do it!')
    window_spike = widgets.IntText(value=spike_window,
                                    description='Window:', disabled=False,
                                    color='black')
    threshold = widgets.FloatText(value=spike_threshold,
                                    description='Threshold:', disabled=False,
                                    color='black')
    influence = widgets.FloatSlider(
        value=spike_influence,
        min=0,
        max=1,
        step=0.1,
        description='Influence:',
        disabled=False,
        continuous_update=False,
        orientation='horizontal',
        readout=True,
        readout_format='.1f',
        slider_color='white'
    )

    # Replace
    replace_label = widgets.Label('Replace QC:')
    replace_checkbox = widgets.Checkbox(description='Do it!')

    # Button
    button = widgets.Button(description='Run tests')
    out = widgets.Output()
    def on_button_clicked(_):
        # "linking function with output"
        with out:
            # what happens when we press the button
            clear_output()
            print('Please wait')
            
            if parameter in wf.parameters:
                show_result(wf, parameter, 
                            'Values without apply any QC test')
                print()

                if reset_checkbox.value:
                    wf.qc_replace(parameters=[parameter], to_replace=1, value=0)
                    wf.qc_replace(parameters=[parameter], to_replace=4, value=0)
                if flat_checkbox.value:
                    wf.qc_flat_test(window=window_flat.value,
                                    parameters=[parameter])
                    show_result(wf, parameter, 
                                'Values after apply the flat test. ' + \
                                ' Configuration flat test: window = ' + \
                                f'{window_flat.value}')
                    print()
                if range_checkbox.value:
                    wf.qc_range_test(limits=limits.value,
                                     parameters=[parameter])
                    show_result(wf, parameter,
                                'Values after apply the flat test and ' + \
                                    'the range test. Configuration ' + \
                                    f'range test: limits = {limits.value}')
                    print()
                if spike_checkbox.value:
                    wf.qc_spike_test(window=window_spike.value,
                                        influence=influence.value,
                                        threshold=threshold.value,
                                        parameters=[parameter])
                    show_result(wf, parameter,
                                'Values after apply the flat test, the ' \
                                    'range test and the spike test.' + \
                                    ' Configuration spike test: window ' \
                                    f'= {window_spike.value}, ' \
                                    f'influence = {influence.value}, ' + \
                                    f'threshold = {threshold.value}')
                    print()
                if replace_checkbox.value:
                    start = 0
                    if spike_checkbox.value:
                        start = window_spike.value
                    wf.qc_replace(parameters=[parameter], start=start)
                    show_result(wf, parameter, f'Final result')
            else:
                print(f'Parameter {parameter} not in {wf.parameters}')

    # linking button and function together using a button's method
    button.on_click(on_button_clicked)

    reset_box = widgets.HBox([reset_label, reset_checkbox])
    flat_box = widgets.HBox([flat_label, flat_checkbox, window_flat])
    range_box = widgets.HBox([range_label, range_checkbox, limits])
    spike_column = widgets.VBox([window_spike, threshold, influence])
    spike_box = widgets.HBox([spike_label, spike_checkbox, spike_column])
    replace_box = widgets.HBox([replace_label, replace_checkbox])

    main_box = widgets.VBox([reset_box, flat_box, range_box, spike_box,
                             replace_box, button, out])

    return main_box



""" Implementation of WaterFrame.reduce_memory() """
import numpy as np


def reduce_memory(self, inplace=True):
    """
    Data in a WaterFrame is saved in a pandas DataFrame. By default, the obtained dataset can be
    very heavy in memory because column types of the pandas DataFrame are defined to accept heavy
    value types. This method redefine the column types ad typically reduced the dataset size in
    memory by 30% - 50%.
    Ref: https://www.kaggle.com/arjanso/reducing-dataframe-memory-size-by-65

    Parameters
    ----------
        inplace: bool
            If false, the method will return a new WaterFrame. Otherwise, it returns True
    """

    def reduce_mem_usage(props):

        NAlist = [] # Keeps track of columns that have missing values filled in. 
        for col in props.columns:
            if props[col].dtype != object:  # Exclude strings
                
                # make variables for Int, max and min
                IsInt = False
                mx = props[col].max()
                mn = props[col].min()
                    
                # test if column can be converted to an integer
                asint = props[col].fillna(0).astype(np.int64)
                result = (props[col] - asint)
                result = result.sum()
                if result > -0.01 and result < 0.01:
                    IsInt = True

                try:
                    # Make Integer/unsigned Integer datatypes
                    if IsInt:
                        if mn >= 0:
                            if mx < 255:
                                props[col] = props[col].astype(np.uint8)
                            elif mx < 65535:
                                props[col] = props[col].astype(np.uint16)
                            elif mx < 4294967295:
                                props[col] = props[col].astype(np.uint32)
                            else:
                                props[col] = props[col].astype(np.uint64)
                        else:
                            if mn > np.iinfo(np.int8).min and mx < np.iinfo(np.int8).max:
                                props[col] = props[col].astype(np.int8)
                            elif mn > np.iinfo(np.int16).min and mx < np.iinfo(np.int16).max:
                                props[col] = props[col].astype(np.int16)
                            elif mn > np.iinfo(np.int32).min and mx < np.iinfo(np.int32).max:
                                props[col] = props[col].astype(np.int32)
                            elif mn > np.iinfo(np.int64).min and mx < np.iinfo(np.int64).max:
                                props[col] = props[col].astype(np.int64)    
                    
                    # Make float datatypes 32 bit
                    else:
                        props[col] = props[col].astype(np.float32)
                except ValueError:
                    # There are nans
                    props[col] = props[col].astype(np.float32)

        return props, NAlist

    df = self.data.copy()
    df, _ = reduce_mem_usage(df)

    if inplace:
        self.data = df.copy()
    else:
        wf = self.copy()
        wf.data = df.copy()
        return wf



""" Implementation of WaterFrame.qc_flat_test(parameters=None, window=2, flag=4) """


def qc_flat_test(self, parameters=None, window=3, flag=4, inplace=True):
    """
    It detects if there are equal consecutive values in the time series.

    Parameters
    ----------
        parameters: string or list of strings, optional
        (parameters = None)
            Parameter to apply the test.
        window: int, optional (window = 3)
            Size of the moving window of values to calculate the mean.
            If it is 0, the function calculates the optimal window.
        flag: int, optional (flag = 4)
            Flag value to write in on the fail test values.
        inplace: bool
            If True, it changes the flags in place and returns True.
            Otherwhise it returns an other WaterFrame.

    Returns
    -------
        new_wf: WaterFrame
    """
    if parameters is None:
        parameters = self.parameters
    elif isinstance(parameters, str):
        parameters = [parameters]

    if window == 0:
        window = 2

    df_rolling = self.data.rolling(window).std()

    data = self.data.copy()

    for parameter in parameters:
        if '_QC' in parameter:
            return False
        else:
            data.loc[df_rolling[parameter] == 0, parameter + '_QC'] = flag

    if inplace:
        self.data = data
        return True
    else:
        new_wf = self.copy()
        new_wf.data = data
        return new_wf



import json
from elasticsearch import Elasticsearch, exceptions
from time import sleep


def to_es(self, data_index_name='data', metadata_index_name='metadata',
          summary_index_name='summary', vocabulary_index_name='vocabulary',
          qc_to_ingest=[0, 1], parameters=None, metadata_to_es=True,
          data_to_es=True, summary_to_es=True, vocabulary_to_es=True, start=None,
          **kwargs):
    """
    Injestion of the WaterFrame into a ElasticSeach DB.

    Parameters
    ----------
        data_index_name: str
            Name of the ElasticSearch index that contains the WaterFrame.data documents.
        metadata_index_name: str
            Name of the ElasticSearch index that contains the WaterFrame.metadata documents.
        summary_index_name: str
            Name of the ElasticSearch index that contains the summary documents.
        vocabulary_index_name: str
            Name of the ElasticSearch index that contains the summary documentation.
        qc_to_intest: list of int
            QC Flags of data to be ingested to the ElasticSearch DB.
        parameters: list of str
            List of parameters to ingest. If parameters is None, all parameters will be ingested.
        metadata_to_es: bool
            If True, metadata will be ingested.
        data_to_es: bool
            If True, data will be ingested.
        vocabulary_to_es: bool
            If True, vocabulary will be ingested.
        **kwargs: Elasticsearch object creation arguments.
            See https://elasticsearch-py.readthedocs.io/en/master/index.html#
    """
    def data_ingestion(es, data_index_name, list_data_dict, start_in):

        bulk_size = len(list_data_dict)
        try:
            print(list_data_dict[1]['parameter'], "to ingest:", bulk_size, "records")
        except IndexError:
            pass
        while bulk_size >= 100:
            if start_in:
                if bulk_size <= start_in:
                    try:
                        es.bulk(index=data_index_name, body=list_data_dict[:100], refresh=True)
                    except ValueError:
                        # Empty body
                        pass
                    list_data_dict = list_data_dict[100:]
                    bulk_size -= 100
                else:
                    list_data_dict = list_data_dict[start_in:]
                    bulk_size = start_in
            else:
                try:
                    es.bulk(index=data_index_name, body=list_data_dict[:100], refresh=True)
                except ValueError:
                    # Empty body
                    pass
                list_data_dict = list_data_dict[100:]
                bulk_size -= 100
            try:
                print(list_data_dict[1]['parameter'], "to ingest:", bulk_size, "records")
            except IndexError:
                pass
        if bulk_size > 0:
            es.bulk(index=data_index_name, body=list_data_dict, refresh=True)

        return True

    def summary_ingestion(es, summary_index_name, summary_dict):

        try:
            # Read emso-stats
            response = es.get(index=summary_index_name, id='datalab')
            es_summary = response['_source']

            es_summary['sites'] = list(set(es_summary['sites'] + summary_dict['sites']))

            es_summary['networks'] = list(set(es_summary['networks'] + summary_dict['networks']))
            param_list = []

            for stats_parameter in summary_dict['parameters']:
                tengui = False
                for summary_parameter in es_summary['parameters']:
                    if summary_parameter['acronym'] == stats_parameter['acronym'] and \
                        'units' in stats_parameter:
                        # print('tengui')
                        tengui = True
                        break
                if not tengui:
                    param_list.append(stats_parameter)
            es_summary['parameters'] += param_list

            # Ingest updated stats
            es.index(index=summary_index_name, id='datalab', body=es_summary)
            return True

        except exceptions.NotFoundError:
            summary_json = json.dumps(summary_dict)
            es.index(index=summary_index_name, id='datalab', body=summary_json)
            return True

    def get_metadata(es, metadata_index_name, search):
        """
        Get a list with the metadata information.
        """

        search_body = {
            'query': {
                'bool': {
                    'must': []
                }
            }  
        }
        if search.get('platform_code'):
            search_body['query']['bool']['must'].append(
                {'term': {'platform_code': search.get('platform_code')}})
        if search.get('geospatial_lat_min'):
            search_body['query']['bool']['must'].append(
                {'term': {'geospatial_lat_min': search.get('geospatial_lat_min')}})
        if search.get('geospatial_lon_min'):
            search_body['query']['bool']['must'].append(
                {'term': {'geospatial_lon_min': search.get('geospatial_lon_min')}})

        # Search query in metadata index
        response = es.search(index=metadata_index_name, body=search_body)
        
        if response['hits']['hits']:
            # Create a metadata dict with the metadata info and the id of the index
            for hit in response['hits']['hits']:
                return hit['_id'], hit['_source']['time_coverage_start'], hit['_source']['time_coverage_end']
        else:
            return False

    def metadata_ingestion(es, metadata_index_name, metadata_dict):

        metadata_search = {
            'platform_code': metadata_dict['platform_code'],
            'time_coverage_start': metadata_dict['time_coverage_start'],
            'time_coverage_end': metadata_dict['time_coverage_end']
        }

        result = get_metadata(es, metadata_index_name, metadata_search)

        if result:
            metadata_id = result[0]
            start = result[1]
            end = result[2]

            metadata_dict['id'] = metadata_id
            if metadata_dict.get('time_coverage_start') > start:
                metadata_dict['time_coverage_start'] = start
            if metadata_dict.get('time_coverage_end') < end:
                metadata_dict['time_coverage_end'] = end

        metadata_json = json.dumps(metadata_dict)
        # Create or update document
        try:
            es.delete(index=metadata_index_name, id=metadata_dict['id'])
        except exceptions.NotFoundError:
            pass
        es.index(index=metadata_index_name, id=metadata_dict['id'], body=metadata_json)
        return True
    
    es = Elasticsearch(**kwargs)

    if metadata_to_es:
        # Add parameters in metadata information
        parameters_string = []
        for parameter in self.parameters:
            parameters_string.append(
                f"{parameter} - {self.vocabulary[parameter]['long_name']} [{self.vocabulary[parameter]['units']}]")

        self.metadata['parameters'] = parameters_string

        ok = metadata_ingestion(es, metadata_index_name, self.metadata)
        if ok:
            print(f"Metadata {self.metadata['id']} ingested")
        else:
            print(f"Error: Metadata {self.metadata['id']} not ingested")
    
    if summary_to_es:
        summary_dict = {
            'sites': [self.metadata['site']],
            'networks': [self.metadata['network']],
            'parameters': []}
        for key, value in self.vocabulary.items():
            if '_QC' not in key or 'TIME' not in key or 'DEPH' not in key:
                try:
                    summary_dict['parameters'].append(
                        {
                            'long_name': value['long_name'],
                            'acronym': key,
                            'units': value['units']})
                except KeyError:
                    print(f"Caution: {key} doesn't contain 'long_name'")

        ok = summary_ingestion(es, summary_index_name, summary_dict)
        if ok:
            print('Stats updated')
        else:
            print('Error: Stats not updated')

    if data_to_es:
        if parameters is None:
            parameters = self.parameters

        for parameter in self.parameters:

            print(f'Parameter to ingest: {parameter}')
            print(f'Total parameters [{self.parameters}]')
            # Only upload the input parameters
            if parameter not in parameters:
                continue

            mini_wf = self.copy()
            try:
                mini_wf.data = self.data[[parameter, parameter+'_QC', 'DEPH', 'DEPH_QC', 'TIME_QC']]
            except KeyError:
                mini_wf.data = self.data[[parameter, parameter+'_QC', 'TIME_QC']].copy()
                mini_wf.data['DEPH'] = mini_wf.data.index.get_level_values('DEPTH')
                if 'DEPTH_QC' in self.data.keys():
                    mini_wf.data['DEPH_QC'] = self.data['DEPTH_QC']
                else:
                    mini_wf.data['DEPH_QC'] = 0

            mini_wf.data = mini_wf.data.dropna()

            mini_wf.data.reset_index(inplace=True)
            mini_wf.data.set_index('TIME', inplace=True)

            df_init = mini_wf.data

            for qc_value in qc_to_ingest:
                df = df_init[df_init[parameter+'_QC'] == qc_value]

                list_data_dict = []
                for index, row in df.iterrows():

                    index_dict = {
                        'index': {
                            # '_id': f"{self.metadata['id']}_{parameter}_{str(index)[:19].replace(' ', 'T')}_{str(row['DEPH'])}"
                            '_id': f"{self.metadata['id']}_{parameter}_{str(index).replace(' ', 'T')}_{str(row['DEPH'])}"
                        }
                    }
                    list_data_dict.append(index_dict)

                    data_dict = {
                        'parameter': parameter,
                        'time': str(index)[:19],
                        'time_qc': int(row['TIME_QC']),
                        'depth': str(row['DEPH']),
                        'depth_qc': int(row['DEPH_QC']),
                        'value': str(row[parameter]),
                        'value_qc': int(row[parameter+'_QC']),
                        'metadata_id': self.metadata['id'],
                        'platform_code': self.metadata['platform_code'],
                        'institution': self.metadata['institution'],
                        'area': self.metadata['area'],
                        'long_name': self.vocabulary[parameter]['long_name'],
                        'units': self.vocabulary[parameter]['units'],
                        'location': {
                            'lat': self.metadata['last_latitude_observation'],
                            'lon': self.metadata['last_longitude_observation'],
                        },
                        'location_qc': 0,
                    }
                    list_data_dict.append(data_dict)

                if list_data_dict:
                    ok = data_ingestion(es, data_index_name, list_data_dict, start)
                    if ok:
                        print(f"Data from {parameter} with QC {qc_value} ingested")
                    else:
                        print(f"ERROR in ingestion of data from {parameter}")

                    # Wait for ES operations
                    sleep(0.01)

            print(f'Parameter ingested: {parameter}')
            print(f'Total parameters [{self.parameters}]')

    if vocabulary_to_es:
        for key, value in self.vocabulary.items():
            if '_QC' in key or key in [
                'DEPTH', 'LATITUDE', 'LONGITUDE', 'TIME', 'station_name'] :
                continue

            vocabulary_dict = {
                'acronim': key,
                'long_name': value['long_name'],
                'units':  value['units']
            }

            vocabulary_json = json.dumps(vocabulary_dict)
            # Create or update document
            es.index(index=vocabulary_index_name, id=key, body=vocabulary_json)



import pandas as pd
import plotly.express as px
import plotly.graph_objects as go


def iplot_data_intervals(self, resample_rule='D', **kwds):
    """
    It uses plotly.express.timeline.
    It creates a plot to view the time intervals of the parameters.

    Parameters
    ----------
        resample_rule: None, str
            Resample rule.
        **kwds: Arguments
            Arguments from plotly.express.timeline
    Returns
    -------
        fig: plotly.graph_objects.Figure
    """
    if resample_rule:
        _wf = self.resample(rule=resample_rule, inplace=False)
    else:
        _wf = self.copy()

    df_content = []
    for parameter in _wf.parameters:
        intervals = _wf.time_intervals(parameter, resample_rule)
        
        for start, end in intervals:
            df_content.append(
                dict(
                    Task=_wf.vocabulary[parameter].get('long_name', parameter),
                    Start=start,
                    Finish=end,
                    Resource=parameter))
    df = pd.DataFrame(df_content)

    fig = px.timeline(
        df,
        x_start="Start",
        x_end="Finish",
        y="Task",
        color="Resource",
        title='Data intervals',
        labels={'Task': 'Parameters'})

    fig.update(layout_showlegend=False)
    fig.update_layout(margin=dict(l=250, r=0, t=30, b=0))

    return fig
